#define _FAR_NO_NAMELESS_UNIONS
#define _FAR_USE_FARFINDDATA
#include "plugin.hpp"
#include <stdio.h>

#define USE_PFLAGS_PANELLEFT
#define IDENTIFY_EDITOR_BY_FILENAME
  // Because identify by EditorID doesn't work - I don't know the ID of my editor
#define OVERWRITE_DIALOG_COMPLEX

#define LOG_MESSAGE(mode,...) { FILE * ltpHandle = fopen("ltp.log", mode); fprintf(ltpHandle, __VA_ARGS__); fclose(ltpHandle); }
#define SHOW_MESSAGE(...) { const char *MsgItems[] = { "Debug", __VA_ARGS__ }; Info.Message(Info.ModuleNumber, FMSG_MB_OK, NULL, MsgItems, sizeof(MsgItems)/sizeof(MsgItems[0]), 1); }

#ifdef __GNUC__
#define _i64(num) num##ll
#define _ui64(num) num##ull
#else
#define _i64(num) num##i64
#define _ui64(num) num##ui64
#endif

/****************************************************************************
 *     startup-    GCC
 ****************************************************************************/
#if defined(__GNUC__)
#include "crt.hpp"

#ifdef __cplusplus
extern "C"{
#endif
//  BOOL WINAPI DllMainCRTStartup(HANDLE hDll, DWORD dwReason, LPVOID lpReserved);
#ifdef __cplusplus
};
#endif

BOOL WINAPI DllMainCRTStartup(HANDLE hDll, DWORD dwReason, LPVOID lpReserved)
{
  (void) lpReserved;
  (void) dwReason;
  (void) hDll;
  return TRUE;
}
#endif

/****************************************************************************
 *      .lng 
 ****************************************************************************/
enum CompareLng {
  MNoLngStringDefined = -1,

  MOK,
  MCancel,

  MCompare,

  MCmpTitle,
  MProcessBox,
  MProcessSubfolders,
  MUseMaxScanDepth,
  MProcessSelected,
  MCompareBox,
  MCompareTime,
  MCompareLowPrecision,
  MCompareIgnoreTimeZone,
  MCompareSize,
  MCompareContents,
  MCompareContentsIgnore,
  MCompareIgnoreNewLines,
  MCompareIgnoreWhitespace,
  MMessageWhenNoDiff,

  MFilePanelsRequired,

  MComparing,
  MComparingWith,

  MComparingFiles,

  MNoDiffTitle,
  MNoDiffBody,

  MNoMemTitle,
  MNoMemBody,

  MOldFARTitle,
  MOldFARBody,

  MEscTitle,
  MEscBody,
  
  MShowReport,
  MSynchronizeDirs,
  
  MCopying,
  MCopyingTo,
  
  MWarning,
  MLeftFileAlreadyExists,
  MRightFileAlreadyExists,
  MNew,
  MExisting,
  MOverwrite,
  MOverwriteAll,
  MSkip,
  MSkipAll,
  
  MFailedToOverwriteFile,
  
	MCopyRightToLeft,
  MOverwriteRightToLeft,
	MCopyLeftToRight,
  MOverwriteLeftToRight,
  
  MReportFileTitle,
};

/****************************************************************************
 *   
 ****************************************************************************/
struct Options {
  int ProcessSubfolders,
      UseMaxScanDepth,
      MaxScanDepth,
      ProcessSelected,
      ProcessHidden,
      CompareTime,
      LowPrecisionTime,
      IgnorePossibleTimeZoneDifferences,
      CompareSize,
      CompareContents,
      CompareContentsIgnore,
      IgnoreWhitespace,
      IgnoreNewLines,
      MessageWhenNoDiff,
      ShowReport;
} Opt;

struct SyncOptions {
	int CopyLeft,
		CopyRight,
		AskOverwriteLeft,
		AskOverwriteRight;
};

struct PendingSync {
	PendingSync * next;
#ifndef IDENTIFY_EDITOR_BY_FILENAME
	int EditorID;
#endif
	char LeftDir[NM];
	char RightDir[NM];
	char ReportFileName[NM];
};

static struct PendingSync * psEditorsQueue = NULL;

/****************************************************************************
 *    FAR
 ****************************************************************************/
static struct PluginStartupInfo Info;
static struct FarStandardFunctions FSF;
static char  *PluginRootKey = NULL;

static void WFD2FFD(WIN32_FIND_DATA &wfd, FAR_FIND_DATA &ffd)
{
  ffd.dwFileAttributes = wfd.dwFileAttributes;
  ffd.ftCreationTime   = wfd.ftCreationTime;
  ffd.ftLastAccessTime = wfd.ftLastAccessTime;
  ffd.ftLastWriteTime  = wfd.ftLastWriteTime;
  ffd.nFileSizeHigh    = wfd.nFileSizeHigh;
  ffd.nFileSizeLow     = wfd.nFileSizeLow;
  ffd.dwReserved0      = wfd.dwReserved0;
  ffd.dwReserved1      = wfd.dwReserved1;

  lstrcpy(ffd.cFileName, wfd.cFileName);
  lstrcpy(ffd.cAlternateFileName, wfd.cAlternateFileName);
}

/****************************************************************************
 *    FAR:    .lng-
 ****************************************************************************/
static const char *GetMsg(int CompareLng)
{
  return Info.GetMsg(Info.ModuleNumber, CompareLng);
}


static int iTruncLen;

/****************************************************************************
 *      (   )
 *      
 ****************************************************************************/
static void TrunCopy(char *cpDest, const char *cpSrc, const char cFiller = ' ')
{
  int iLen = lstrlen(FSF.TruncStr(lstrcpy(cpDest, cpSrc), iTruncLen));

  if (iLen < iTruncLen)
  {
    memset(&cpDest[iLen], cFiller, iTruncLen - iLen);
    cpDest[iTruncLen] = '\0';
  }
}

static bool bStart;

/****************************************************************************
 *      
 ****************************************************************************/
static void ShowMessage(const char *Name1, const char *Name2, FARINT64 Progress, FARINT64 Max, CompareLng msg1 = MComparing, CompareLng msg2 = MComparingWith)
{
  static DWORD dwTicks;
  DWORD dwNewTicks = GetTickCount();
  if (dwNewTicks - dwTicks < 500)
    return;
  dwTicks = dwNewTicks;

  char TruncName1[NM], TruncName2[NM], ProgressBar[NM];
  TrunCopy(TruncName1, Name1);
  TrunCopy(TruncName2, Name2);

  __int64 iProgress;
  if (Max.i64 <= 0)
    iProgress = 0;
  else if (Progress.i64 < Max.i64)
    iProgress = iTruncLen * Progress.i64 / Max.i64;
  else
    iProgress = iTruncLen;
  ProgressBar[iTruncLen] = '\0';
  memset(&ProgressBar[0], '\xdb', iProgress);
  memset(&ProgressBar[iProgress], '\xb0', iTruncLen - iProgress);

  const char *MsgItems[] = {
    GetMsg(MCmpTitle),
    GetMsg(msg1),
    TruncName1,
    GetMsg(msg2),
    TruncName2,
    ProgressBar
  };
  Info.Message(Info.ModuleNumber, bStart ? FMSG_LEFTALIGN :
               FMSG_LEFTALIGN|FMSG_KEEPBACKGROUND,
               NULL,
               MsgItems,
               sizeof(MsgItems) / sizeof(MsgItems[0]),
               0);
  bStart = false;
}


/****************************************************************************
 *    ShowDialog
 ****************************************************************************/
long WINAPI ShowDialogProc(HANDLE hDlg, int Msg, int Param1, long Param2)
{
  static int CompareContents,
             CompareContentsIgnore,
             ProcessSubfolders,
             CompareTime;

  switch (Msg)
  {
    case DN_INITDIALOG:
      CompareContents = ((DWORD)Info.SendDlgMessage(hDlg, DM_GETDLGDATA, 0, 0)) & 0x000000FF;
      CompareContentsIgnore = CompareContents + 1;
      ProcessSubfolders = (((DWORD)Info.SendDlgMessage(hDlg, DM_GETDLGDATA, 0, 0)) >> 8) & 0x000000FF;
      CompareTime = ((DWORD)Info.SendDlgMessage(hDlg, DM_GETDLGDATA, 0, 0)) >> 16;
      break;

    case DN_BTNCLICK:
      if (Param1 == CompareTime || Param1 == ProcessSubfolders || Param1 == CompareContents || Param1 == CompareContentsIgnore)
      {
        if (Param2)
        {
          Info.SendDlgMessage(hDlg, DM_ENABLE, Param1+1, TRUE);
          if (!(Param1 == CompareContents && !Info.SendDlgMessage(hDlg, DM_GETCHECK, CompareContentsIgnore, 0)))
          {
            Info.SendDlgMessage(hDlg, DM_ENABLE, Param1+2, TRUE);
            if (Param1 == CompareContents)
              Info.SendDlgMessage(hDlg, DM_ENABLE, Param1+3, TRUE);
          }
        }
        else
        {
          Info.SendDlgMessage(hDlg, DM_ENABLE, Param1+1, FALSE);
          Info.SendDlgMessage(hDlg, DM_ENABLE, Param1+2, FALSE);
          if (Param1 == CompareContents)
            Info.SendDlgMessage(hDlg, DM_ENABLE, Param1+3, FALSE);
        }
      }
      break;
  }

  return Info.DefDlgProc(hDlg, Msg, Param1, Param2);
}

/****************************************************************************
 *    ,     ,
 *   Opt,  ( )    ,
 *  true,    OK
 ****************************************************************************/
static bool ShowDialog(bool bPluginPanels, bool bSelectionPresent)
{
  static struct InitDialogItem {
    unsigned char Type;
    unsigned char X1, Y1, X2, Y2;
    int           Data;
    int           DefaultRegValue;
    char         *SelectedRegValue;
    unsigned int  Flags;
    int          *StoreTo;
  } InitItems[] = {
    /* 0*/ { DI_DOUBLEBOX,    3,  1, 62, 20, MCmpTitle,                0, NULL,                                   0, NULL },
    /* 1*/ { DI_TEXT,         5,  2,  0,  0, MProcessBox,              0, NULL,                                   0, NULL },
    /* 2*/ { DI_CHECKBOX,     5,  3,  0,  0, MProcessSubfolders,       0, "ProcessSubfolders",                    0, &Opt.ProcessSubfolders },
    /* 3*/ { DI_CHECKBOX,     9,  4,  0,  0, MUseMaxScanDepth,         0, "UseMaxScanDepth",                      0, &Opt.UseMaxScanDepth },
    /* 4*/ { DI_FIXEDIT,      0,  4,  4,  0, MNoLngStringDefined,     99, "MaxScanDepth",              DIF_MASKEDIT, &Opt.MaxScanDepth },
    /* 5*/ { DI_CHECKBOX,     5,  5,  0,  0, MProcessSelected,         0, "ProcessSelected",                      0, &Opt.ProcessSelected },
    /* 6*/ { DI_TEXT,         0,  6,  0,  0, MNoLngStringDefined,      0, NULL,                       DIF_SEPARATOR, NULL },
    /* 7*/ { DI_TEXT,         5,  7,  0,  0, MCompareBox,              0, NULL,                                   0, NULL },
    /* 8*/ { DI_CHECKBOX,     5,  8,  0,  0, MCompareTime,             1, "CompareTime",                          0, &Opt.CompareTime },
    /* 9*/ { DI_CHECKBOX,     9,  9,  0,  0, MCompareLowPrecision,     1, "LowPrecisionTime",                     0, &Opt.LowPrecisionTime },
    /*10*/ { DI_CHECKBOX,     9, 10,  0,  0, MCompareIgnoreTimeZone,   1, "IgnorePossibleTimeZoneDifferences",    0, &Opt.IgnorePossibleTimeZoneDifferences },
    /*11*/ { DI_CHECKBOX,     5, 11,  0,  0, MCompareSize,             1, "CompareSize",                          0, &Opt.CompareSize },
    /*12*/ { DI_CHECKBOX,     5, 12,  0,  0, MCompareContents,         0, "CompareContents",                      0, &Opt.CompareContents },
    /*13*/ { DI_CHECKBOX,     9, 13,  0,  0, MCompareContentsIgnore,   0, "CompareContentsIgnore",                0, &Opt.CompareContentsIgnore },
    /*14*/ { DI_RADIOBUTTON, 13, 14,  0,  0, MCompareIgnoreNewLines,   1, "IgnoreNewLines",               DIF_GROUP, &Opt.IgnoreNewLines },
    /*15*/ { DI_RADIOBUTTON, 13, 15,  0,  0, MCompareIgnoreWhitespace, 0, "IgnoreWhitespace",                     0, &Opt.IgnoreWhitespace },
    /*16*/ { DI_TEXT,         0, 16,  0,  0, MNoLngStringDefined,      0, NULL,                       DIF_SEPARATOR, NULL },
    /*17*/ { DI_CHECKBOX,     5, 17,  0,  0, MShowReport,              1, "ShowReport",                           0, &Opt.ShowReport },
    /*18*/ { DI_CHECKBOX,     5, 18,  0,  0, MMessageWhenNoDiff,       0, "MessageWhenNoDiff",                    0, &Opt.MessageWhenNoDiff },
    /*19*/ { DI_TEXT,         0, 19,  0,  0, MNoLngStringDefined,      0, NULL,                       DIF_SEPARATOR, NULL },
    /*20*/ { DI_BUTTON,       0, 20,  0,  0, MOK,                      0, NULL,                     DIF_CENTERGROUP, NULL },
    /*21*/ { DI_BUTTON,       0, 20,  0,  0, MCancel,                  0, NULL,                     DIF_CENTERGROUP, NULL }
  };
  struct FarDialogItem DialogItems[sizeof(InitItems) / sizeof(InitItems[0])];
  char Mask[] = "99999";

  memset(DialogItems,0,sizeof(DialogItems));

  HKEY hKey;
  if (!PluginRootKey ||
       RegOpenKeyEx(HKEY_CURRENT_USER, PluginRootKey, 0, KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS)
    hKey = 0;

  int DlgData=0;
  bool bNoFocus = true;
  size_t i;
  for (i = 0; i < sizeof(InitItems) / sizeof(InitItems[0]); i++)
  {
    DWORD dwRegValue;
    DWORD dwSize                  = sizeof(DWORD);
    DialogItems[i].Type           = InitItems[i].Type;
    DialogItems[i].X1             = InitItems[i].X1;
    DialogItems[i].Y1             = InitItems[i].Y1;
    DialogItems[i].X2             = InitItems[i].X2;
    DialogItems[i].Y2             = InitItems[i].Y2;
    DialogItems[i].Focus          = FALSE;
    DialogItems[i].Flags          = InitItems[i].Flags;
    lstrcpy(DialogItems[i].Data.Data, (InitItems[i].Data == MNoLngStringDefined)
            ? "" : GetMsg(InitItems[i].Data));

    dwRegValue = (hKey && InitItems[i].SelectedRegValue &&
                  RegQueryValueEx(hKey, InitItems[i].SelectedRegValue, NULL,
                  NULL, (unsigned char *)&dwRegValue, &dwSize)
                  == ERROR_SUCCESS ) ? dwRegValue : InitItems[i].DefaultRegValue;
    if (DialogItems[i].Type == DI_CHECKBOX || DialogItems[i].Type == DI_RADIOBUTTON)
      DialogItems[i].Param.Selected = dwRegValue;
    else if (DialogItems[i].Type == DI_FIXEDIT)
    {
      FSF.itoa(dwRegValue, DialogItems[i].Data.Data, 10);
      DialogItems[i].Param.Mask = Mask;
      DialogItems[i].X1 = DialogItems[i-1].X1 + lstrlen(DialogItems[i-1].Data.Data)
                          - (strchr(DialogItems[i-1].Data.Data, '&')?1:0) + 5;
      DialogItems[i].X2 += DialogItems[i].X1;
    }

    switch (InitItems[i].Data)
    {
      case MCompareContents:
        DlgData += i;
        if (bPluginPanels)
        {
          DialogItems[i].Flags |= DIF_DISABLE;
          DialogItems[i].Param.Selected = 0;
        }
        if (!DialogItems[i].Param.Selected)
        {
          InitItems[i+1].Flags |= DIF_DISABLE;
          InitItems[i+2].Flags |= DIF_DISABLE;
          InitItems[i+3].Flags |= DIF_DISABLE;
        }
        else
        {
          InitItems[i+1].Flags &= ~DIF_DISABLE;
          InitItems[i+2].Flags &= ~DIF_DISABLE;
          InitItems[i+3].Flags &= ~DIF_DISABLE;
        }
        break;
      case MCompareContentsIgnore:
        if (!DialogItems[i].Param.Selected || DialogItems[i].Flags & DIF_DISABLE)
        {
          InitItems[i+1].Flags |= DIF_DISABLE;
          InitItems[i+2].Flags |= DIF_DISABLE;
        }
        else
        {
          InitItems[i+1].Flags &= ~DIF_DISABLE;
          InitItems[i+2].Flags &= ~DIF_DISABLE;
        }
        break;
      case MCompareIgnoreWhitespace:
        if (DialogItems[i].Param.Selected == DialogItems[i-1].Param.Selected)
        {
          DialogItems[i-1].Param.Selected = 1;
          DialogItems[i].Param.Selected = 0;
        }
        break;
      case MProcessSubfolders:
        DlgData += i<<8;
        if (bPluginPanels)
        {
          DialogItems[i].Flags |= DIF_DISABLE;
          DialogItems[i].Param.Selected = 0;
        }
        if (!DialogItems[i].Param.Selected)
        {
          InitItems[i+1].Flags |= DIF_DISABLE;
          InitItems[i+2].Flags |= DIF_DISABLE;
        }
        else
        {
          InitItems[i+1].Flags &= ~DIF_DISABLE;
          InitItems[i+2].Flags &= ~DIF_DISABLE;
        }
        break;
      case MProcessSelected:
        if (!bSelectionPresent)
        {
          DialogItems[i].Flags |= DIF_DISABLE;
          DialogItems[i].Param.Selected = 0;
        }
        break;
      case MCompareTime:
        DlgData += i<<16;
        if (!DialogItems[i].Param.Selected)
        {
          InitItems[i+1].Flags |= DIF_DISABLE;
          InitItems[i+2].Flags |= DIF_DISABLE;
        }
        else
        {
          InitItems[i+1].Flags &= ~DIF_DISABLE;
          InitItems[i+2].Flags &= ~DIF_DISABLE;
        }
        break;
      case MOK:
        DialogItems[i].DefaultButton = 1;
        break;
    }

    if (bNoFocus && DialogItems[i].Type == DI_CHECKBOX && !(DialogItems[i].Flags & DIF_DISABLE))
    {
      DialogItems[i].Focus = TRUE;
      bNoFocus = false;
    }
  }

  if (hKey)
    RegCloseKey(hKey);

  int ExitCode = Info.DialogEx(Info.ModuleNumber, -1, -1, 66, 22, "Contents",
                               DialogItems, sizeof(DialogItems) / sizeof(DialogItems[0]),
                               0, 0, ShowDialogProc, DlgData);
  if (ExitCode == (sizeof(InitItems) / sizeof(InitItems[0]) - 1) || ExitCode == -1)
    return false;

  for (i = 0; i < sizeof(InitItems) / sizeof(InitItems[0]); i++)
    if (InitItems[i].StoreTo)
      if (InitItems[i].Type == DI_CHECKBOX || InitItems[i].Type == DI_RADIOBUTTON)
        *InitItems[i].StoreTo = (DWORD)DialogItems[i].Param.Selected;
      else if (InitItems[i].Type == DI_FIXEDIT)
        *InitItems[i].StoreTo = FSF.atoi(DialogItems[i].Data.Data);

  DWORD dwDisposition;
  if (PluginRootKey &&
       RegCreateKeyEx(HKEY_CURRENT_USER, PluginRootKey, 0, NULL, REG_OPTION_NON_VOLATILE,
       KEY_ALL_ACCESS, NULL, &hKey, &dwDisposition) == ERROR_SUCCESS)
  {
    for (i = 0; i < sizeof(InitItems) / sizeof(InitItems[0]); i++)
      if (!(DialogItems[i].Flags & DIF_DISABLE) && InitItems[i].SelectedRegValue)
      {
        DWORD dwValue = *InitItems[i].StoreTo;
        RegSetValueEx(hKey, InitItems[i].SelectedRegValue, 0, REG_DWORD, (BYTE *)&dwValue, sizeof(dwValue));
      }
    RegCloseKey(hKey);
  }

  if (bPluginPanels)
  {
    Opt.ProcessSubfolders = FALSE;
    Opt.CompareContents = FALSE;
  }
  Opt.ProcessHidden = (Info.AdvControl(Info.ModuleNumber, ACTL_GETPANELSETTINGS, NULL) &
                        FPS_SHOWHIDDENANDSYSTEMFILES) != 0;

  return true;
}

static bool bBrokenByEsc;
static HANDLE hConInp = INVALID_HANDLE_VALUE;

/****************************************************************************
 *   Esc.  true,    Esc
 ****************************************************************************/
static bool CheckForEsc(void)
{
  if (hConInp == INVALID_HANDLE_VALUE)
    return false;

  static DWORD dwTicks;
  DWORD dwNewTicks = GetTickCount();
  if (dwNewTicks - dwTicks < 500)
    return false;
  dwTicks = dwNewTicks;

  INPUT_RECORD rec;
  DWORD ReadCount;
  while (PeekConsoleInput(hConInp, &rec, 1, &ReadCount) && ReadCount)
  {
    ReadConsoleInput(hConInp, &rec, 1, &ReadCount);
    if ( rec.EventType == KEY_EVENT && rec.Event.KeyEvent.wVirtualKeyCode == VK_ESCAPE &&
         rec.Event.KeyEvent.bKeyDown )
      //     Esc
      if ( Info.AdvControl(Info.ModuleNumber, ACTL_GETCONFIRMATIONS, NULL) & FCS_INTERRUPTOPERATION )
      {
        const char *MsgItems[] = {
          GetMsg(MEscTitle),
          GetMsg(MEscBody),
          GetMsg(MOK),
          GetMsg(MCancel)
        };
        if ( !Info.Message(Info.ModuleNumber,
                           FMSG_WARNING,
                           NULL,
                           MsgItems,
                           sizeof(MsgItems) / sizeof(MsgItems[0]),
                           2) )
          return bBrokenByEsc = true;
      }
      else
        return bBrokenByEsc = true;
  }

  return false;
}

/****************************************************************************
 *        
 ****************************************************************************/
static char *BuildFullFilename(const char *cpDir, const char *cpFileName)
{
  static char cName[NM];
  FSF.AddEndSlash(lstrcpy(cName, cpDir));

  return lstrcat(cName, cpFileName);
}

struct FileIndex {
  PluginPanelItem **ppi;
  int iCount;
};

/****************************************************************************
 *        PluginPanelItem
 *   qsort()
 ****************************************************************************/
static int __cdecl PICompare(const void *el1, const void *el2)
{
  const PluginPanelItem *ppi1 = *(const PluginPanelItem **)el1, *ppi2 = *(const PluginPanelItem **)el2;

  if (ppi1->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
  {
    if (!(ppi2->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
      return -1;
  }
  else
  {
    if (ppi2->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      return 1;
  }

  return -FSF.LStricmp(ppi1->FindData.cFileName, ppi2->FindData.cFileName);
}

/****************************************************************************
 *       
 ****************************************************************************/
static bool BuildPanelIndex(const struct PanelInfo *pInfo, struct FileIndex *pIndex)
{
  bool bProcessSelected;
  pIndex->ppi = NULL;
  pIndex->iCount = ( bProcessSelected = (Opt.ProcessSelected && pInfo->SelectedItemsNumber &&
                     (pInfo->SelectedItems[0].Flags & PPIF_SELECTED)) ) ? pInfo->SelectedItemsNumber :
                     pInfo->ItemsNumber;
  if (!pIndex->iCount)
    return true;
  if (!(pIndex->ppi = (PluginPanelItem **)malloc(pIndex->iCount * sizeof(pIndex->ppi[0]))))
    return false;
  int j = 0;
  for (int i = pInfo->ItemsNumber - 1; i >= 0 && j < pIndex->iCount; i--)
    if ( (Opt.ProcessSubfolders ||
         !(pInfo->PanelItems[i].FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) &&
         (!bProcessSelected || (pInfo->PanelItems[i].Flags & PPIF_SELECTED)) &&
         lstrcmp(pInfo->PanelItems[i].FindData.cFileName, "..") &&
         lstrcmp(pInfo->PanelItems[i].FindData.cFileName, ".") )
      {
        pIndex->ppi[j++] = &pInfo->PanelItems[i];
      }
  if (pIndex->iCount = j)
    FSF.qsort(pIndex->ppi, j, sizeof(pIndex->ppi[0]), PICompare);
  else
  {
    free(pIndex->ppi);
    pIndex->ppi = NULL;
  }

  return true;
}

/****************************************************************************
 *  
 ****************************************************************************/
static void FreePanelIndex(struct FileIndex *pIndex)
{
  if (pIndex->ppi)
    free(pIndex->ppi);
  pIndex->ppi = NULL;
  pIndex->iCount = 0;
}

/****************************************************************************
 *    Info.GetDirList().     
 *      Dir,  .
 ****************************************************************************/
static int GetDirList(const char *Dir, struct PluginPanelItem **pPanelItem, int *pItemsNumber)
{
  *pPanelItem = NULL;
  *pItemsNumber = 0;
  char cPathMask[MAX_PATH];
  WIN32_FIND_DATA wfdFindData;
  HANDLE hFind;
  if ( (hFind = FindFirstFile(lstrcat(lstrcpy(cPathMask, Dir), "\\*"), &wfdFindData)) ==
       INVALID_HANDLE_VALUE )
    return TRUE;

  int iRet = TRUE;
  do
  {
    if (!lstrcmp(wfdFindData.cFileName, ".") || !lstrcmp(wfdFindData.cFileName, ".."))
      continue;
    if ((wfdFindData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) && !Opt.ProcessHidden)
      continue;
    struct PluginPanelItem *pPPI;
    if (!(pPPI = (struct PluginPanelItem *)realloc(*pPanelItem, (*pItemsNumber + 1) * sizeof(*pPPI))))
    {
      free(*pPanelItem);
      *pItemsNumber = 0;
      iRet = FALSE;
      break;
    }
    *pPanelItem = pPPI;
    WFD2FFD(wfdFindData,(*pPanelItem)[(*pItemsNumber)++].FindData);
  } while (FindNextFile(hFind, &wfdFindData));
  FindClose(hFind);
  return iRet;
}

/****************************************************************************
 *    Info.FreeDirList().
 ****************************************************************************/
static void FreeDirList(struct PluginPanelItem *PanelItem)
{
  if (PanelItem)
    free(PanelItem);
}

static bool CompareDirs(const struct PanelInfo *AInfo, const struct PanelInfo *PInfo, 
                        bool bCompareAll, int ScanDepth, 
                        FILE * fpResults, int APanelRootDirLength, bool bActiveIsLeft );
static DWORD bufSize;
static char *ABuf, *PBuf;

bool isnewline(int c)
{
  return (c == '\r' || c == '\n');
}

static __int64 BuildInt64(DWORD lowWord, DWORD highWord) {
       __int64 Result = highWord;
       Result = (Result << 32) | lowWord;
       return Result;
}

static char strMissingLeft[] = "<<==";
static char strMissingRight[] = "==>>";
static char strNewerLeft[] = "--->";
static char strNewerRight[] = "<---";
static char strCantDecide[] = "<-->";
       
static void WriteFileComparisonResultToFile( 
                                             FILE * fpResults, 
                                             const char *CurDir, int RootDirLength, 
                                             const char *FileName, bool bDirectories,
                                             FILETIME *ATime, FILETIME *PTime,
                                             __int64 ASize, __int64 PSize,
                                             bool bActiveIsLeft
                                             )
{
    if (fpResults == 0 || fpResults == INVALID_HANDLE_VALUE)
		return;
    // Adresare me zajimaji jen v pripade, ze se lisi kompletne (na jedne strane adresar je, 
    // na  druhe neni - jinak se ty zmeny promitnou v souborech resp. jsou nepodstatne.
	char szDirection[5];
	// Tvrdy presun doleva: Pokud neni aktivni soubor a aktivni je levy, nebo pokud neni pasivni soubor a aktivni je pravy
	if ((!ATime && bActiveIsLeft) || (!PTime && !bActiveIsLeft))
      strcpy(szDirection, strMissingLeft);
    // Tvrdy presun doprava: Pokud neni aktivni soubor a aktivni je pravy, nebo pokud neni pasivni soubor a aktivni je levy
	else if ((!ATime && !bActiveIsLeft) || (!PTime && bActiveIsLeft))
      strcpy(szDirection, strMissingRight);
    else {
      // Pokud se adresar dostane sem, tak je pro me nezajimavy - znamena to, ze uvnitr toho
      // adresare jsou rozdily, ale pokud na nich zalezi, tak se tak jako tak projevi v rozdilech
      // souboru
      if (bDirectories)
        return;
      bool bAktivniNovejsi = false;
      bool bStejneStare = false;
      if (ATime->dwHighDateTime > PTime->dwHighDateTime)
        bAktivniNovejsi = true;
  	  else if (ATime->dwHighDateTime < PTime->dwHighDateTime)
        bAktivniNovejsi = false;
	  else if (ATime->dwLowDateTime > PTime->dwLowDateTime)
	    bAktivniNovejsi = true;
	  else if (ATime->dwLowDateTime < PTime->dwLowDateTime)
	    bAktivniNovejsi = false;
      else
        bStejneStare = true;
      // Nelze rozhodnout o smeru
      if (bStejneStare)
        strcpy(szDirection, strCantDecide);
      // Mekky presun doprava: Pokud je aktivni soubor novejsi a aktivni je levy, nebo pokud je pasivni soubor novejsi a aktivni je pravy
      else if ((bAktivniNovejsi && bActiveIsLeft) || (!bAktivniNovejsi && !bActiveIsLeft))
        strcpy(szDirection, strNewerLeft);
      // Mekky presun doleva: Pokud je aktivni soubor novejsi a aktivni je pravy, nebo pokud je pasivni soubor novejsi a aktivni je levy
      else if ((bAktivniNovejsi && !bActiveIsLeft) || (!bAktivniNovejsi && bActiveIsLeft))
        strcpy(szDirection, strNewerRight);
      // Sem by to nikdy nemelo dojit
      else
        strcpy(szDirection, "????");
    }
	char szASize[128], szPSize[128];
	FSF.itoa64(ASize, szASize, 10);
	FSF.itoa64(PSize, szPSize, 10);
	if (CurDir[RootDirLength])
	  if (bDirectories) 
  	    fprintf(fpResults, "%s\t%s\\%s\\*\n", szDirection, &CurDir[RootDirLength+1], FileName);
      else
	    fprintf(fpResults, "%s\t%s\\%s\n", szDirection, &CurDir[RootDirLength+1], FileName);
	else
	  if (bDirectories)
  	    fprintf(fpResults, "%s\t%s\\*\n", szDirection, FileName);
	  else
	    fprintf(fpResults, "%s\t%s\n", szDirection, FileName);
}

static int FileExists(const char * Filename, FARINT64 * pSize = NULL) {
	int result = 0;
	HANDLE h = CreateFile(Filename, 0, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (h != INVALID_HANDLE_VALUE) {
		if (pSize != NULL) {
			DWORD hp;
			pSize->Part.LowPart = GetFileSize(h, &hp);
			pSize->Part.HighPart = hp;
			if ((pSize->Part.LowPart == 0xffffffff) && (GetLastError != NO_ERROR))
				pSize->i64 = -1;
		}
		CloseHandle(h);
		result = 1;
	}
	return result;
}

struct SynchronizeFileCopyCallbackData {
	char * srcFileName;
	char * destFileName;
};

DWORD WINAPI SynchronizeFileCopyCallback(
		LARGE_INTEGER TotalFileSize,
    LARGE_INTEGER TotalBytesTransferred,
    LARGE_INTEGER StreamSize,
    LARGE_INTEGER StreamBytesTransferred,
    DWORD dwStreamNumber,
    DWORD dwCallbackReason,
    HANDLE hSourceFile,
    HANDLE hDestinationFile,
    LPVOID lpData)
{
	FARINT64 progress, max;
	progress.i64 = TotalBytesTransferred.QuadPart;
	max.i64 = TotalFileSize.QuadPart;
	if (lpData) {
		struct SynchronizeFileCopyCallbackData * data = (SynchronizeFileCopyCallbackData *)lpData;
		ShowMessage(data->srcFileName, data->destFileName, progress, max, MCopying, MCopyingTo);
	} else {
		ShowMessage("", "", progress, max);
	}
	if (CheckForEsc())
		return PROGRESS_CANCEL;
	else
		return PROGRESS_CONTINUE;
}

static char * FormatFileSize(const char * filename) {
	static char result[65];
	char num[65];
	FARINT64 sz;
	strcpy(result, "");
	if (FileExists(filename, &sz)) {
		if (sz.i64 >= 0) {
#ifdef FORMAT_INT_BY_LOCALE
			FSF.itoa64(sz.i64, num, 10);
			if (!GetNumberFormat(LOCALE_USER_DEFAULT, 0, num, NULL, result, sizeof(result)))
				strcpy(result, "");
#else
			FSF.itoa64(sz.i64, result, 10);
#endif
		}
	}
	return result;
}

static char * FormatFileDateTime(const char * filename) {
	static char result[65];
	char cDate[32], cTime[32];
	HANDLE handle;
	FILETIME fileTime;
	SYSTEMTIME systemTime;
	strcpy(result, "");
	handle = CreateFile(filename, 0, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (handle != INVALID_HANDLE_VALUE) {
		if (GetFileTime(handle, NULL, NULL, &fileTime)) {
			if (FileTimeToSystemTime(&fileTime, &systemTime)) {
				if (GetDateFormat(LOCALE_USER_DEFAULT, 0, &systemTime, "dd.MM.yyyy", cDate, sizeof(cDate))) {
					if (GetTimeFormat(LOCALE_USER_DEFAULT, 0, &systemTime, "HH:mm:ss", cTime, sizeof(cTime))) {
						sprintf(result, "%s %s", cDate, cTime);
					}
				}
			}
		}
		CloseHandle(handle);
	}
	return result;
}

enum QueryOverwriteFileResult {
	QOF_ABORT,
	QOF_OVERWRITE,
	QOF_OVERWRITEALL,
	QOF_SKIP,
	QOF_SKIPALL
};

static QueryOverwriteFileResult QueryOverwriteFile(const int MovingLeft, const char * strSourceFilename, const char * strDestinationFilename, const char * DisplayFilename = NULL)
{
	char cNew[66], cExisting[66], cFilename[NM];
	strncpy(cFilename, (DisplayFilename ? DisplayFilename : strDestinationFilename), sizeof(cFilename));
	cFilename[sizeof(cFilename)-1] = 0;
	FSF.TruncPathStr(cFilename, 66);
	sprintf(cNew, "%-30.30s %15.15s %19.19s", GetMsg(MNew), FormatFileSize(strSourceFilename), FormatFileDateTime(strSourceFilename));
	sprintf(cExisting, "%-30.30s %15.15s %19.19s", GetMsg(MExisting), FormatFileSize(strDestinationFilename), FormatFileDateTime(strDestinationFilename));
#ifdef OVERWRITE_DIALOG_COMPLEX
	struct InitDialogItem {
	  int Type;
	  int X1;
	  int Y1;
	  int X2;
	  int Y2;
	  int Focus;
	  DWORD_PTR Selected;
	  unsigned int Flags;
	  int DefaultButton;
	  char *Data;
	} InitItems[] = {
		//       Type       x1  y1  x2  y2 Foc Sel Flg Def Txt
    /*  0 */ DI_DOUBLEBOX,  3,  1, 72,  9, 0, 0, 0,                 0, (char *)GetMsg(MWarning),
    /*  1 */ DI_TEXT,       5,  2, 70,  2, 0, 0, DIF_CENTERTEXT,    0, (char *)GetMsg(MovingLeft ? MLeftFileAlreadyExists : MRightFileAlreadyExists),
    /*  2 */ DI_TEXT,       5,  3, 70,  2, 0, 0, DIF_SHOWAMPERSAND, 0, cFilename,
    /*  3 */ DI_TEXT,       4,  4, 71,  3, 0, 0, DIF_SEPARATOR,     0, "",
    /*  4 */ DI_TEXT,       5,  5, 70,  4, 0, 0, DIF_SHOWAMPERSAND, 0, cNew,
    /*  5 */ DI_TEXT,       5,  6, 70,  5, 0, 0, DIF_SHOWAMPERSAND, 0, cExisting,
    /*  6 */ DI_TEXT,       4,  7, 71,  6, 0, 0, DIF_SEPARATOR,     0, "",
    /*  7 */ DI_BUTTON,     0,  8,  0,  8, 1, 0, DIF_CENTERGROUP,   0, (char *)GetMsg(MOverwrite),
    /*  8 */ DI_BUTTON,     0,  8,  0,  8, 0, 0, DIF_CENTERGROUP,   0, (char *)GetMsg(MOverwriteAll),
    /*  9 */ DI_BUTTON,     0,  8,  0,  8, 0, 0, DIF_CENTERGROUP,   0, (char *)GetMsg(MSkip),
    /* 10 */ DI_BUTTON,     0,  8,  0,  8, 0, 0, DIF_CENTERGROUP,   0, (char *)GetMsg(MSkipAll),
    /* 11 */ DI_BUTTON,     0,  8,  0,  8, 0, 0, DIF_CENTERGROUP,   0, (char *)GetMsg(MCancel)
  };
  enum DialogResults {
		DR_OVERWRITE = 7,
		DR_OVERWRITE_ALL = 8,
		DR_SKIP = 9,
		DR_SKIP_ALL = 10,
	};
  struct FarDialogItem DialogItems[sizeof(InitItems)/sizeof(InitItems[0])];
  struct InitDialogItem *ii = InitItems;
  struct FarDialogItem *di = DialogItems;
  for (int i = 0; i < sizeof(InitItems)/sizeof(InitItems[0]); i++, ii++, di++) {
		di->Type = ii->Type;
		di->X1 = ii->X1;
		di->Y1 = ii->Y1;
		di->X2 = ii->X2;
		di->Y2 = ii->Y2;
		di->Focus = ii->Focus;
		di->Param.Selected = ii->Selected;
		di->Flags = ii->Flags;
		di->DefaultButton = ii->DefaultButton;
		strcpy(di->Data.Data, ii->Data);
	}
  int ExitCode = Info.DialogEx(Info.ModuleNumber, -1, -1, 76, 11, NULL, DialogItems, sizeof(DialogItems) / sizeof(DialogItems[0]), 0, FDLG_WARNING, NULL, 0);
#else
  const char * MessageLines[] = {
		GetMsg(MWarning),
		GetMsg(MovingLeft ? MLeftFileAlreadyExists : MRightFileAlreadyExists),
		cFilename,
		"",
		cNew,
		cExisting,
		"",
		GetMsg(MOverwrite),
		GetMsg(MOverwriteAll),
		GetMsg(MSkip),
		GetMsg(MSkipAll),
		GetMsg(MCancel)
	};
  enum DialogResults {
		DR_OVERWRITE = 0,
		DR_OVERWRITE_ALL = 1,
		DR_SKIP = 2,
		DR_SKIP_ALL = 3,
	};
  int ExitCode = Info.Message(Info.ModuleNumber, FMSG_WARNING|FMSG_LEFTALIGN, NULL, MessageLines, sizeof(MessageLines) / sizeof(MessageLines[0]), 5);
#endif
  QueryOverwriteFileResult result = QOF_ABORT;
  switch (ExitCode) {
		case DR_OVERWRITE:
			result = QOF_OVERWRITE;
			break;
		case DR_OVERWRITE_ALL:
			result = QOF_OVERWRITEALL;
			break;
		case DR_SKIP:
			result = QOF_SKIP;
			break;
		case DR_SKIP_ALL:
			result = QOF_SKIPALL;
			break;
	}
	return result;
}

static int SynchronizeFile(const char * srcFileName, const char * destFileName, struct SyncOptions * options, int direction, const char * displayFileName = NULL)
{
	int ok = 1;
	FARINT64 progress, max;
	progress.i64 = 0;
	max.i64 = 1;
	if (srcFileName && destFileName && FileExists(srcFileName, &max)) {
		// Zobrazim, co se to vlastne kopiruje
		ShowMessage(srcFileName, destFileName, progress, max, MCopying, MCopyingTo);
		// Pokud existuje cilovy soubor, tak se mozna budu muset zeptat na prepsani
		int doCopy = 1;
		if (FileExists(destFileName))
			if (((direction < 0) && options->AskOverwriteLeft) || ((direction > 0) && options->AskOverwriteRight)) {
				switch (QueryOverwriteFile(direction < 0, srcFileName, destFileName, displayFileName)) {
					case QOF_OVERWRITE:
						doCopy = 1;
						break;
					case QOF_OVERWRITEALL:
						doCopy = 1;
						if (direction < 0)
							options->AskOverwriteLeft = 0;
						else
							options->AskOverwriteRight = 0;
						break;
					case QOF_SKIP:
						doCopy = 0;
						break;
					case QOF_SKIPALL:
						doCopy = 0;
						if (direction < 0)
							options->CopyLeft = 0;
						else
							options->CopyRight = 0;
						break;
					default:
						doCopy = 0;
						ok = 0;
						break;
				}
			}
		if (doCopy) {
			struct SynchronizeFileCopyCallbackData copyData;
			copyData.srcFileName = (char *)srcFileName;
			copyData.destFileName = (char *)destFileName;
			if (!CopyFileEx(srcFileName, destFileName, SynchronizeFileCopyCallback, &copyData, NULL, 0)) {
			  const char * MessageLines[] = {
					GetMsg(MWarning),
					GetMsg(MFailedToOverwriteFile),
					destFileName
				};
			  switch (Info.Message(Info.ModuleNumber, FMSG_WARNING|FMSG_MB_OKCANCEL, NULL, MessageLines, sizeof(MessageLines) / sizeof(MessageLines[0]), 0)) {
					case 0: // OK
						break;
					default:
						ok = 0;
						break;
				}
			}
		}
	}
	return ok;
}

static int SynchronizeDirectory(const char * srcDirName, const char * destDirName, struct SyncOptions * options, int direction)
{
	// Na vstupu ocekavam zdrojovy a cilovy adresar bez zaverecneho lomitka
	// Pokud cilovy adresar neexistuje, vytvorim ho
//SHOW_MESSAGE("Synchronize directory", srcDirName, destDirName);
	WIN32_FIND_DATA fnd;
	HANDLE handle = FindFirstFile(destDirName, &fnd);
	if (handle == INVALID_HANDLE_VALUE) {
		if (!CreateDirectory(destDirName, NULL))
			return 0;
	} else {
		FindClose(handle);
		if (!(fnd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
			return 0;
	}
	// Ted budu rekurzivne prochazet vsechny soubory v adresari
	int result = 1;
	char WildCard[NM];
	snprintf(WildCard, sizeof(WildCard), "%s\\*.*", srcDirName);
	if (handle = FindFirstFile(WildCard, &fnd)) {
		do {
			char NewSrc[NM], NewDest[NM];
			snprintf(NewSrc, sizeof(NewSrc), "%s\\%s", srcDirName, fnd.cFileName);
			snprintf(NewDest, sizeof(NewDest), "%s\\%s", destDirName, fnd.cFileName);
//SHOW_MESSAGE("Synchronize directory LOOP", "Wildcard", WildCard, "Found", fnd.cFileName, "New names", NewSrc, NewDest);
			if (fnd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
				if ((stricmp(fnd.cFileName, ".") != 0) && (stricmp(fnd.cFileName, "..") != 0)) {
					if (!SynchronizeDirectory(NewSrc, NewDest, options, direction)) {
						result = 0;
					}
				}
			} else {
				if (!SynchronizeFile(NewSrc, NewDest, options, direction)) {
					result = 0;
				}
			}
		} while (result && FindNextFile(handle, &fnd));
		FindClose(handle);	
	}
	return result;
}

static void Synchronize(const char * LeftPath, const char * RightPath, const char * ReportFileName, struct SyncOptions * options)
{
  HANDLE hScreen = Info.SaveScreen(0, 0, -1, -1);
  hConInp = CreateFile("CONIN$", GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
  FILE * fpResults = fopen(ReportFileName, "r");
  if (strlen(LeftPath) && strlen(RightPath) && fpResults) {
		char ReportLine[2*NM];
		int ok = 1;
		bStart = true;
		while (ok && fgets(ReportLine, sizeof(ReportLine), fpResults)) {
			int rlLen = strlen(ReportLine);
			if (ReportLine[rlLen-1] = '\n')
				ReportLine[rlLen-1] = 0;
			// Pokusim se synchronizovat soubor dany timto radkem
			int dir = 0;
			int len = 0;
			if (strncmp(ReportLine, strMissingLeft, len = strlen(strMissingLeft)) == 0)
				dir = -1;
			else if (strncmp(ReportLine, strMissingRight, len = strlen(strMissingRight)) == 0)
			  dir = 1;
			else if (strncmp(ReportLine, strNewerRight, len = strlen(strNewerRight)) == 0)
				dir = -1;
			else if (strncmp(ReportLine, strNewerLeft, len = strlen(strNewerLeft)) == 0)
				dir = 1;
			if (dir != 0) {
				if (((dir < 0) && options->CopyLeft) || ((dir > 0) && options->CopyRight)) {
					char * p = &(ReportLine[len]);
					while (*p && ((*p == ' ') || (*p == '\t')))
						p++;
					if (*p) {
						// Muze nastat specialni pripad, pokud radek konci "\\*" - to znaci, ze 
						// se dany adresar na druhem panelu vubec nevyskytuje a je treba ho zkopirovat
						// komplet cely. Jinak se kopiruje jen jeden soubor.
						len = strlen(p);
						int directory = (p[len-1] == '*');
						if (directory) {
							p[len-1] = 0;
							if ((len > 2) && (p[len-2] == '\\'))
								p[len-2] = 0;
						}
						char srcFileName[NM+1], destFileName[NM+1];
						strncpy(srcFileName, BuildFullFilename((dir < 0) ? RightPath : LeftPath, p), sizeof(srcFileName));
						strncpy(destFileName, BuildFullFilename((dir < 0) ? LeftPath : RightPath, p), sizeof(destFileName));
						if (directory) {
							if (!SynchronizeDirectory(srcFileName, destFileName, options, dir))
								ok = 0;
						} else {
							if (!SynchronizeFile(srcFileName, destFileName, options, dir, p))
								ok = 0;
						}
					}
				}
			}
		}
	if (fpResults)
		fclose(fpResults);
	}
  DeleteFile(ReportFileName);
  CloseHandle(hConInp);
  Info.RestoreScreen(hScreen);
}

static int GetSyncOptions(struct SyncOptions * options, const char * ReportFileName) {
	options->CopyLeft = 0;
	options->CopyRight = 0;
	options->AskOverwriteLeft = ( Info.AdvControl(Info.ModuleNumber, ACTL_GETCONFIRMATIONS, NULL) & FCS_COPYOVERWRITE );
	options->AskOverwriteRight = ( Info.AdvControl(Info.ModuleNumber, ACTL_GETCONFIRMATIONS, NULL) & FCS_COPYOVERWRITE );
  FILE * fpResults = fopen(ReportFileName, "r");
  if (fpResults) {
		char ReportLine[2*NM];
		while ((!(options->CopyLeft && options->CopyRight)) && fgets(ReportLine, sizeof(ReportLine), fpResults)) {
			if (strncmp(ReportLine, strMissingLeft, strlen(strMissingLeft)) == 0)
				options->CopyLeft = 1;
			else if (strncmp(ReportLine, strMissingRight, strlen(strMissingRight)) == 0)
				options->CopyRight = 1;
			else if (strncmp(ReportLine, strNewerRight, strlen(strNewerRight)) == 0)
				options->CopyLeft = 1;
			else if (strncmp(ReportLine, strNewerLeft, strlen(strNewerLeft)) == 0)
				options->CopyRight = 1;
		}
		fclose(fpResults);
	}
	if (options->CopyLeft || options->CopyRight) {
		// Jsou nejake soubory na synchronizaci. Takze si udelam dialog, kde se zeptam,
		// co vlastne chci delat.
		struct InitDialogItem {
		  int Type;
		  int X1;
		  int Y1;
		  int X2;
		  int Y2;
		  int Focus;
		  DWORD_PTR Selected;
		  unsigned int Flags;
		  int DefaultButton;
		  char *Data;
		} InitItems[] = {
			//       Type           x1  y1  x2  y2 Foc Sel               Flg                                            Def Txt
	    /*  0 */ DI_DOUBLEBOX,  3,  1, 42,  9, 0, 0,                          0,                                     0, (char *)GetMsg(MCmpTitle),
	    /*  1 */ DI_CHECKBOX,   5,  2, 40,  2, 1, options->CopyLeft,          options->CopyLeft  ? 0 : DIF_DISABLE,  0, (char *)GetMsg(MCopyRightToLeft),
	    /*  2 */ DI_CHECKBOX,   8,  3, 40,  3, 0, options->AskOverwriteLeft,  options->CopyLeft  ? 0 : DIF_DISABLE,  0, (char *)GetMsg(MOverwriteRightToLeft),
	    /*  3 */ DI_CHECKBOX,   5,  4, 40,  4, 0, options->CopyRight,         options->CopyRight ? 0 : DIF_DISABLE,  0, (char *)GetMsg(MCopyLeftToRight),
	    /*  4 */ DI_CHECKBOX,   8,  5, 40,  5, 0, options->AskOverwriteRight, options->CopyRight ? 0 : DIF_DISABLE,  0, (char *)GetMsg(MOverwriteLeftToRight),
	    /*  5 */ DI_TEXT,       4,  6, 41,  6, 0, 0, DIF_SEPARATOR,     0, "",
	    /*  6 */ DI_BUTTON,     0,  7,  0,  7, 0, 0, DIF_CENTERGROUP,   1, (char *)GetMsg(MOK),
	    /*  7 */ DI_BUTTON,     0,  7,  0,  7, 0, 0, DIF_CENTERGROUP,   0, (char *)GetMsg(MCancel)
	  };
	  enum DialogResults {
			DR_CB_COPYLEFT = 1,
			DR_CB_OVERWRITELEFT = 2,
			DR_CB_COPYRIGHT = 3,
			DR_CB_OVERWRITERIGHT = 4,
			DR_OK = 6,
		};
	  struct FarDialogItem DialogItems[sizeof(InitItems)/sizeof(InitItems[0])];
	  struct InitDialogItem *ii = InitItems;
	  struct FarDialogItem *di = DialogItems;
	  for (int i = 0; i < sizeof(InitItems)/sizeof(InitItems[0]); i++, ii++, di++) {
			di->Type = ii->Type;
			di->X1 = ii->X1;
			di->Y1 = ii->Y1;
			di->X2 = ii->X2;
			di->Y2 = ii->Y2;
			di->Focus = ii->Focus;
			di->Param.Selected = ii->Selected;
			di->Flags = ii->Flags;
			di->DefaultButton = ii->DefaultButton;
			strcpy(di->Data.Data, ii->Data);
		}
	  int ExitCode = Info.DialogEx(Info.ModuleNumber, -1, -1, 46, 11, NULL, DialogItems, sizeof(DialogItems) / sizeof(DialogItems[0]), 0, 0, NULL, 0);
		if (ExitCode == DR_OK) {
			options->CopyLeft = DialogItems[DR_CB_COPYLEFT].Param.Selected;
			options->AskOverwriteLeft = DialogItems[DR_CB_OVERWRITELEFT].Param.Selected;
			options->CopyRight = DialogItems[DR_CB_COPYRIGHT].Param.Selected;
			options->AskOverwriteRight = DialogItems[DR_CB_OVERWRITERIGHT].Param.Selected;
			return 1;
		}
	}
	return 0;
}


/****************************************************************************
 *         ( 
 * ).
 *  true,   .
 ****************************************************************************/
static bool CompareFiles( const FAR_FIND_DATA *AData, const FAR_FIND_DATA *PData,
                          const char *ACurDir, const char *PCurDir, int ScanDepth,
						  FILE * fpResults, int APanelRootDirLength, bool bActiveIsLeft )
{
  if (AData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
  {
    //    
    if (Opt.ProcessSubfolders)
    {
      if (Opt.UseMaxScanDepth && Opt.MaxScanDepth<ScanDepth+1)
        return true;
      //     
      struct PanelInfo AInfo, PInfo;
      memset(&AInfo, 0, sizeof(AInfo));
      memset(&PInfo, 0, sizeof(PInfo));
      bool bEqual;
      if (!GetDirList(lstrcpy(AInfo.CurDir, BuildFullFilename(ACurDir, AData->cFileName)),
                              &AInfo.PanelItems, &AInfo.ItemsNumber)
          || !GetDirList(lstrcpy(PInfo.CurDir, BuildFullFilename(PCurDir, PData->cFileName)),
                              &PInfo.PanelItems, &PInfo.ItemsNumber))
      {
        bBrokenByEsc = true; //    ,    
        bEqual = false; //  
      }
      else
        bEqual = CompareDirs(&AInfo, &PInfo, fpResults, ScanDepth+1, fpResults, APanelRootDirLength, bActiveIsLeft);
      FreeDirList(AInfo.PanelItems);
      FreeDirList(PInfo.PanelItems);
      return bEqual;
    }
  }
  else
  {
    //    
    if (Opt.CompareSize)
      if (AData->nFileSizeLow != PData->nFileSizeLow || AData->nFileSizeHigh != PData->nFileSizeHigh)
        return false;
    if (Opt.CompareTime)
    {
      if (Opt.LowPrecisionTime || Opt.IgnorePossibleTimeZoneDifferences)
      {
        union {
          __int64 num;
          struct
          {
            DWORD lo;
            DWORD hi;
          } hilo;
        } Precision, Difference, TimeDelta, temp;

        Precision.hilo.hi = 0;
        Precision.hilo.lo = Opt.LowPrecisionTime ? 20000000 : 0; //2s or 0s
        Difference.num = _i64(9000000000); //15m

        if (AData->ftLastWriteTime.dwHighDateTime > PData->ftLastWriteTime.dwHighDateTime)
        {
          TimeDelta.hilo.hi = AData->ftLastWriteTime.dwHighDateTime - PData->ftLastWriteTime.dwHighDateTime;
          TimeDelta.hilo.lo = AData->ftLastWriteTime.dwLowDateTime - PData->ftLastWriteTime.dwLowDateTime;
          if (TimeDelta.hilo.lo > AData->ftLastWriteTime.dwLowDateTime)
            --TimeDelta.hilo.hi;
        }
        else
        {
          if (AData->ftLastWriteTime.dwHighDateTime == PData->ftLastWriteTime.dwHighDateTime)
          {
            TimeDelta.hilo.hi = 0;
            TimeDelta.hilo.lo = max(PData->ftLastWriteTime.dwLowDateTime,AData->ftLastWriteTime.dwLowDateTime)-
                                min(PData->ftLastWriteTime.dwLowDateTime,AData->ftLastWriteTime.dwLowDateTime);
          }
          else
          {
            TimeDelta.hilo.hi = PData->ftLastWriteTime.dwHighDateTime - AData->ftLastWriteTime.dwHighDateTime;
            TimeDelta.hilo.lo = PData->ftLastWriteTime.dwLowDateTime - AData->ftLastWriteTime.dwLowDateTime;
            if (TimeDelta.hilo.lo > PData->ftLastWriteTime.dwLowDateTime)
              --TimeDelta.hilo.hi;
          }
        }

        //     26 .
        if (Opt.IgnorePossibleTimeZoneDifferences)
        {
          int counter=0;
          while (TimeDelta.hilo.hi > Difference.hilo.hi && counter<=26*4)
          {
            temp.hilo.lo = TimeDelta.hilo.lo - Difference.hilo.lo;
            temp.hilo.hi = TimeDelta.hilo.hi - Difference.hilo.hi;
            if (temp.hilo.lo > TimeDelta.hilo.lo)
              --temp.hilo.hi;
            TimeDelta.hilo.lo = temp.hilo.lo;
            TimeDelta.hilo.hi = temp.hilo.hi;
            ++counter;
          }
          if (counter<=26*4 && TimeDelta.hilo.hi == Difference.hilo.hi)
          {
            TimeDelta.hilo.hi = 0;
            TimeDelta.hilo.lo = max(TimeDelta.hilo.lo,Difference.hilo.lo) - min(TimeDelta.hilo.lo,Difference.hilo.lo);
          }
        }

        if (Precision.hilo.hi < TimeDelta.hilo.hi ||
            (Precision.hilo.hi == TimeDelta.hilo.hi && Precision.hilo.lo < TimeDelta.hilo.lo))
          return false;
      }
      else if (AData->ftLastWriteTime.dwLowDateTime != PData->ftLastWriteTime.dwLowDateTime ||
               AData->ftLastWriteTime.dwHighDateTime != PData->ftLastWriteTime.dwHighDateTime)
        return false;
    }
    if (Opt.CompareContents)
    {
      HANDLE hFileA, hFileP;
      char cpFileA[MAX_PATH], cpFileP[MAX_PATH];
      FARINT64 Progress, Max, TempI64;
      Progress.i64 = 0;
      Max.Part.LowPart = AData->nFileSizeLow;
      Max.Part.HighPart = AData->nFileSizeHigh;
      TempI64.i64 = 0;
      ShowMessage(lstrcpy(cpFileA, BuildFullFilename(ACurDir, AData->cFileName)),
                  lstrcpy(cpFileP, BuildFullFilename(PCurDir, PData->cFileName)),
                  Progress, Max);
      if ((hFileA = CreateFile(cpFileA, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                               OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE)
        return false;
      if ((hFileP = CreateFile(cpFileP, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                               OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE)
      {
        CloseHandle(hFileA);
        return false;
      }

      bool bEqual = true;
      DWORD ReadSizeA, ReadSizeP;
      if (!Opt.CompareContentsIgnore)
      {
        do
        {
          ShowMessage(cpFileA, cpFileP, Progress, Max);
          if (CheckForEsc() || !ReadFile(hFileA, ABuf, bufSize, &ReadSizeA, NULL) ||
              !ReadFile(hFileP, PBuf, bufSize, &ReadSizeP, NULL) || (ReadSizeA != ReadSizeP))
          {
            bEqual = false;
            break;
          }
          if (!ReadSizeA)
            break;
          if (memcmp(ABuf, PBuf, ReadSizeA))
          {
            bEqual = false;
            break;
          }
          TempI64.Part.LowPart = ReadSizeA;
          Progress.i64 += TempI64.i64;
        } while (ReadSizeA == bufSize);
      }
      else
      {
        ReadSizeA=1;
        ReadSizeP=1;
        char *PtrA=ABuf+ReadSizeA, *PtrP=PBuf+ReadSizeP;
        bool bExpectNewLineA = false;
        bool bExpectNewLineP = false;
        while (1)
        {
          while (PtrA >= ABuf+ReadSizeA && ReadSizeA)
          {
            if (CheckForEsc() || !ReadFile(hFileA, ABuf, bufSize, &ReadSizeA, NULL))
            {
              bEqual = false;
              break;
            }
            PtrA=ABuf;
          }

          if (!bEqual)
            break;

          while (PtrP >= PBuf+ReadSizeP && ReadSizeP)
          {
            if (CheckForEsc() || !ReadFile(hFileP, PBuf, bufSize, &ReadSizeP, NULL))
            {
              bEqual = false;
              break;
            }
            PtrP=PBuf;
          }

          if (!bEqual || (!ReadSizeP && !ReadSizeA))
            break;

          if (Opt.IgnoreWhitespace)
          {
            while (PtrA < ABuf+ReadSizeA && PtrP < PBuf+ReadSizeP && !isspace(*PtrA) && !isspace(*PtrP))
            {
              if (*PtrA != *PtrP)
              {
                bEqual = false;
                break;
              }
              ++PtrA;
              ++PtrP;
            }

            if (!bEqual)
              break;

            while (PtrA < ABuf+ReadSizeA && isspace(*PtrA))
              ++PtrA;

            while (PtrP < PBuf+ReadSizeP && isspace(*PtrP))
              ++PtrP;
          }
          else
          {
            if (bExpectNewLineA)
            {
              bExpectNewLineA = false;
              if (PtrA < ABuf+ReadSizeA && *PtrA == '\n')
                ++PtrA;
            }

            if (bExpectNewLineP)
            {
              bExpectNewLineP = false;
              if (PtrP < PBuf+ReadSizeP && *PtrP == '\n')
                ++PtrP;
            }

            while (PtrA < ABuf+ReadSizeA && PtrP < PBuf+ReadSizeP && !isnewline(*PtrA) && !isnewline(*PtrP))
            {
              if (*PtrA != *PtrP)
              {
                bEqual = false;
                break;
              }
              ++PtrA;
              ++PtrP;
            }

            if (!bEqual)
              break;

            if (PtrA < ABuf+ReadSizeA && PtrP < PBuf+ReadSizeP && (!isnewline(*PtrA) || !isnewline(*PtrP)))
            {
              bEqual = false;
              break;
            }

            if (PtrA < ABuf+ReadSizeA && PtrP < PBuf+ReadSizeP)
            {
              if (*PtrA == '\r')
                bExpectNewLineA = true;

              if (*PtrP == '\r')
                bExpectNewLineP = true;

              ++PtrA;
              ++PtrP;
            }
          }

          if (PtrA < ABuf+ReadSizeA && !ReadSizeP)
          {
            bEqual = false;
            break;
          }

          if (PtrP < PBuf+ReadSizeP && !ReadSizeA)
          {
            bEqual = false;
            break;
          }
        }
      }

      CloseHandle(hFileA);
      CloseHandle(hFileP);
      return bEqual;
    }
  }
  return true;
}

/****************************************************************************
 *   ,   AInfo  PInfo.
 *  true,   .
 *  bCompareAll ,
 *        PPIF_SELECTED (bCompareAll == true)
 *    false    (bCompareAll == false).
 ****************************************************************************/
static bool CompareDirs(const struct PanelInfo *AInfo, const struct PanelInfo *PInfo, 
                        bool bCompareAll, int ScanDepth, 
                        FILE * fpResults, int APanelRootDirLength, bool bActiveIsLeft )
{
  //      
  struct FileIndex sfiA, sfiP;
  char DirA[MAX_PATH], DirP[MAX_PATH];
  FARINT64 Progress, Max;
  Progress.i64 = 0;
  Max.i64 = 1;
  ShowMessage(lstrcpy(DirA, BuildFullFilename(AInfo->CurDir, "*")),
              lstrcpy(DirP, BuildFullFilename(PInfo->CurDir, "*")),
              Progress, Max);
  if (!BuildPanelIndex(AInfo, &sfiA) || !BuildPanelIndex(PInfo, &sfiP))
  {
    const char *MsgItems[] = {
      GetMsg(MNoMemTitle),
      GetMsg(MNoMemBody),
      GetMsg(MOK)
    };
    Info.Message(Info.ModuleNumber,
                 FMSG_WARNING,
                 NULL,
                 MsgItems,
                 sizeof(MsgItems) / sizeof(MsgItems[0]),
                 1);
    bBrokenByEsc = true;
    FreePanelIndex(&sfiA);
    FreePanelIndex(&sfiP);
    return true;
  }
  bool bDifferenceNotFound = true;
  int i = sfiA.iCount - 1, j = sfiP.iCount - 1;
  while (i >= 0 && j >= 0 && (bDifferenceNotFound || bCompareAll) && !bBrokenByEsc)
  {
    const int iMaxCounter = 256;
    static int iCounter = iMaxCounter;
    if (!--iCounter)
    {
      iCounter = iMaxCounter;
      if (CheckForEsc())
        break;
    }
    switch (PICompare(&sfiA.ppi[i], &sfiP.ppi[j]))
    {
      case 0: //   -   
        if (CompareFiles(&sfiA.ppi[i]->FindData, &sfiP.ppi[j]->FindData, AInfo->CurDir, PInfo->CurDir, ScanDepth, fpResults, APanelRootDirLength, bActiveIsLeft))
        { //   
          sfiA.ppi[i--]->Flags &= ~PPIF_SELECTED;
          sfiP.ppi[j--]->Flags &= ~PPIF_SELECTED;
        }
        else
        {
          // Soubory se lisi
          __int64 ASize, PSize;
          ASize = BuildInt64(sfiA.ppi[i]->FindData.nFileSizeLow, sfiA.ppi[i]->FindData.nFileSizeHigh);
          PSize = BuildInt64(sfiP.ppi[j]->FindData.nFileSizeLow, sfiP.ppi[j]->FindData.nFileSizeHigh);
          WriteFileComparisonResultToFile(
              fpResults, 
              AInfo->CurDir, APanelRootDirLength,
              sfiA.ppi[i]->FindData.cFileName, (sfiA.ppi[i]->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) && (sfiP.ppi[j]->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY),
              &sfiA.ppi[i]->FindData.ftLastWriteTime, &sfiP.ppi[j]->FindData.ftLastWriteTime,
              ASize, PSize,
              bActiveIsLeft
              );
          bDifferenceNotFound = false;
          sfiA.ppi[i--]->Flags |= PPIF_SELECTED;
          sfiP.ppi[j--]->Flags |= PPIF_SELECTED;
        }
        break;
      case 1: //  sfiA.ppi[i]     sfiP.ppi
        // Pasivni soubor neexistuje
        __int64 ASize;
        ASize = BuildInt64(sfiA.ppi[i]->FindData.nFileSizeLow, sfiA.ppi[i]->FindData.nFileSizeHigh);
        WriteFileComparisonResultToFile(
            fpResults, 
            AInfo->CurDir, APanelRootDirLength,
            sfiA.ppi[i]->FindData.cFileName, sfiA.ppi[i]->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY,
            &sfiA.ppi[i]->FindData.ftLastWriteTime, 0,
            ASize, -1,
            bActiveIsLeft
            );
        bDifferenceNotFound = false;
        sfiA.ppi[i--]->Flags |= PPIF_SELECTED;
        break;
      case -1: //  sfiP.ppi[j]     sfiA.ppi
        // Aktivni soubor neexistuje
        __int64 PSize;
        PSize = BuildInt64(sfiP.ppi[j]->FindData.nFileSizeLow, sfiP.ppi[j]->FindData.nFileSizeHigh);
        WriteFileComparisonResultToFile(
            fpResults, 
            AInfo->CurDir, APanelRootDirLength,
            sfiP.ppi[j]->FindData.cFileName, sfiP.ppi[j]->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY,
            0, &sfiP.ppi[j]->FindData.ftLastWriteTime,
            -1, PSize,
            bActiveIsLeft
            );
        bDifferenceNotFound = false;
        sfiP.ppi[j--]->Flags |= PPIF_SELECTED;
        break;
    }
  }
  if (!bBrokenByEsc)
  {
    //   .  ,     
    if (i >= 0)
    {
      bDifferenceNotFound = false;
      if (bCompareAll)
        for (; i >= 0; i--) {
          // Pasivni soubor neexistuje
          __int64 ASize;
          ASize = BuildInt64(sfiA.ppi[i]->FindData.nFileSizeLow, sfiA.ppi[i]->FindData.nFileSizeHigh);
          WriteFileComparisonResultToFile(
              fpResults, 
              AInfo->CurDir, APanelRootDirLength,
              sfiA.ppi[i]->FindData.cFileName, sfiA.ppi[i]->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY,
              &sfiA.ppi[i]->FindData.ftLastWriteTime, 0,
              ASize, -1,
              bActiveIsLeft
              );
          sfiA.ppi[i]->Flags |= PPIF_SELECTED;
        }
    }
    if (j >= 0)
    {
      bDifferenceNotFound = false;
      if (bCompareAll)
        for (; j >= 0; j--) {
          // Aktivni soubor neexistuje
          __int64 PSize;
          PSize = BuildInt64(sfiP.ppi[j]->FindData.nFileSizeLow, sfiP.ppi[j]->FindData.nFileSizeHigh);
          WriteFileComparisonResultToFile(
              fpResults, 
              AInfo->CurDir, APanelRootDirLength,
              sfiP.ppi[j]->FindData.cFileName, sfiP.ppi[j]->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY,
              0, &sfiP.ppi[j]->FindData.ftLastWriteTime,
              -1, PSize,
              bActiveIsLeft
              );
          sfiP.ppi[j]->Flags |= PPIF_SELECTED;
        }
    }
  }
  FreePanelIndex(&sfiA);
  FreePanelIndex(&sfiP);
  return bDifferenceNotFound;
}

/****************************************************************************
 ***************************** Exported functions ***************************
 ****************************************************************************/

static bool bOldFAR = false;

/****************************************************************************
 *    FAR    
 ****************************************************************************/
void WINAPI _export SetStartupInfo(const struct PluginStartupInfo *Info)
{
  const char *cpPlugRegKey = "\\AdvCompare";
  ::Info = *Info;

  if (Info->StructSize >= (int)sizeof(struct PluginStartupInfo))
    FSF = *Info->FSF;
  else
    bOldFAR = true;

  if (PluginRootKey)
  {
    free(PluginRootKey);
    PluginRootKey = NULL;
  }

  if (PluginRootKey = (char *)malloc(lstrlen(Info->RootKey) + lstrlen(cpPlugRegKey) + 1))
  {
    lstrcpy(PluginRootKey, Info->RootKey);
    lstrcat(PluginRootKey, cpPlugRegKey);
  }
  else
  {
    const char *MsgItems[] = {
      GetMsg(MNoMemTitle),
      GetMsg(MNoMemBody),
      GetMsg(MOK)
    };
    ::Info.Message(::Info.ModuleNumber,
                   FMSG_WARNING,
                   NULL,
                   MsgItems,
                   sizeof(MsgItems) / sizeof(MsgItems[0]),
                   1);
  }
}

/****************************************************************************
 *    FAR    
 ****************************************************************************/
void WINAPI _export GetPluginInfo(struct PluginInfo *Info)
{
  static const char *PluginMenuStrings[1];

  Info->StructSize              = (int)sizeof(*Info);
  Info->Flags                   = 0;
  Info->DiskMenuStrings         = NULL;
  PluginMenuStrings[0]          = GetMsg(MCompare);
  Info->PluginMenuStrings       = PluginMenuStrings;
  Info->PluginMenuStringsNumber = sizeof(PluginMenuStrings) / sizeof(PluginMenuStrings[0]);
  Info->PluginConfigStrings     = NULL;
  Info->CommandPrefix           = NULL;
}

/****************************************************************************
 *   . FAR  ,    
 ****************************************************************************/
HANDLE WINAPI _export OpenPlugin(int OpenFrom, int Item)
{
  //     ...
  if (bOldFAR)
  {
    const char *MsgItems[] = {
      GetMsg(MOldFARTitle),
      GetMsg(MOldFARBody),
      GetMsg(MOK)
    };
    Info.Message(Info.ModuleNumber,
                 FMSG_WARNING,
                 NULL,
                 MsgItems,
                 sizeof(MsgItems) / sizeof(MsgItems[0]),
                 1);

    return INVALID_HANDLE_VALUE;
  }

  struct PanelInfo AInfo, PInfo;

  //       ...
  if ( !Info.Control(INVALID_HANDLE_VALUE, FCTL_GETPANELINFO, &AInfo) ||
       !Info.Control(INVALID_HANDLE_VALUE, FCTL_GETANOTHERPANELINFO, &PInfo) )
  {
    return INVALID_HANDLE_VALUE;
  }

  //   ...
  if (AInfo.PanelType != PTYPE_FILEPANEL || PInfo.PanelType != PTYPE_FILEPANEL)
  {
    const char *MsgItems[] = {
      GetMsg(MCmpTitle),
      GetMsg(MFilePanelsRequired),
      GetMsg(MOK)
    };
    Info.Message(Info.ModuleNumber,
                 FMSG_WARNING,
                 NULL,
                 MsgItems,
                 sizeof(MsgItems) / sizeof(MsgItems[0]),
                 1);

    return INVALID_HANDLE_VALUE;
  }

  // Decide whether active panel is left or right
  bool bActiveIsLeft = true;
#ifdef USE_PFLAGS_PANELLEFT
  if (AInfo.Flags & PFLAGS_PANELLEFT)
    bActiveIsLeft = true;
  else if (PInfo.Flags & PFLAGS_PANELLEFT)
    bActiveIsLeft = false;
#endif

  //      ...
  if ( !ShowDialog(AInfo.Plugin || PInfo.Plugin,
                  (AInfo.SelectedItemsNumber && (AInfo.SelectedItems[0].Flags & PPIF_SELECTED)) ||
                  (PInfo.SelectedItemsNumber && (PInfo.SelectedItems[0].Flags & PPIF_SELECTED))) )
  {
    return INVALID_HANDLE_VALUE;
  }

  //       Esc...
  HANDLE hScreen = Info.SaveScreen(0, 0, -1, -1);
  hConInp = CreateFile("CONIN$", GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);

  //     ...
  HANDLE hConOut = CreateFile("CONOUT$", GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
  CONSOLE_SCREEN_BUFFER_INFO csbiNfo;
  if (GetConsoleScreenBufferInfo(hConOut, &csbiNfo))
  {
    if ((iTruncLen = csbiNfo.dwSize.X - 20) > NM - 2)
      iTruncLen = NM - 2;
    else if (iTruncLen < 0)
      iTruncLen = csbiNfo.dwSize.X - csbiNfo.dwSize.X / 4;
  }
  else
    iTruncLen = 60;
  CloseHandle(hConOut);

  //       ...
  char cConsoleTitle[MAX_PATH], cBuffer[MAX_PATH];
  DWORD dwTitleSaved = GetConsoleTitle(cConsoleTitle, sizeof(cConsoleTitle));
  OSVERSIONINFO ovi;
  ovi.dwOSVersionInfoSize = sizeof(ovi);
  if (GetVersionEx(&ovi) && ovi.dwPlatformId != VER_PLATFORM_WIN32_NT)
    OemToChar(GetMsg(MComparingFiles), cBuffer);
  else
    lstrcpy(cBuffer, GetMsg(MComparingFiles));
  SetConsoleTitle(cBuffer);

  //      ...
  HKEY hKey;
  if ( !PluginRootKey || RegOpenKeyEx(HKEY_CURRENT_USER, PluginRootKey, 0, KEY_QUERY_VALUE, &hKey) !=
       ERROR_SUCCESS )
    hKey = 0;
  DWORD dwSize = sizeof(DWORD);
  bufSize = ( hKey && (RegQueryValueEx(hKey, "CompareBufferSize", NULL, NULL, (unsigned char *)&bufSize,
              &dwSize) == ERROR_SUCCESS) && (bufSize > 32767) ) ? bufSize : 32768;
  if (hKey)
    RegCloseKey(hKey);
  ABuf = (char*)malloc(bufSize);
  PBuf = (char*)malloc(bufSize);

  bBrokenByEsc = false;
  bStart = true;
  bool bDifferenceNotFound = false;
  
  char cResultsFileName[MAX_PATH] = "";
  // char *cLeft = NULL, *cRight = NULL;
  char cLeft[NM] = "", cRight[NM] = "";
  FILE * fpResults = 0; 
  if (FSF.MkTemp(cResultsFileName, 0)) {
  	fpResults = fopen(cResultsFileName, "w");
  	if (bActiveIsLeft) {
			strcpy(cLeft, AInfo.CurDir);
			strcpy(cRight, PInfo.CurDir);
      //cLeft = AInfo.CurDir;
      //cRight = PInfo.CurDir;
    } else {
			strcpy(cLeft, PInfo.CurDir);
			strcpy(cRight, AInfo.CurDir);
      //cLeft = PInfo.CurDir;
      //cRight = AInfo.CurDir;
    }
  	fprintf(fpResults, GetMsg(MReportFileTitle), cLeft, cRight);
  }
  
  //      ...
  if (ABuf && PBuf) {
    int RootDirLength = strlen(AInfo.CurDir);
    if ((RootDirLength > 0) && (AInfo.CurDir[RootDirLength-1] == '\\'))
       RootDirLength--;
    bDifferenceNotFound = CompareDirs(&AInfo, &PInfo, true, 0, fpResults, RootDirLength, bActiveIsLeft);
  }

  if (fpResults)
    fclose(fpResults);
    
  free (ABuf);
  free (PBuf);
  CloseHandle(hConInp);
  Info.RestoreScreen(hScreen);

  //     .    ...
  if (!bBrokenByEsc)
  {
    Info.Control(INVALID_HANDLE_VALUE, FCTL_SETSELECTION, &AInfo);
    Info.Control(INVALID_HANDLE_VALUE, FCTL_SETANOTHERSELECTION, &PInfo);
    Info.Control(INVALID_HANDLE_VALUE, FCTL_REDRAWPANEL, NULL);
    Info.Control(INVALID_HANDLE_VALUE, FCTL_REDRAWANOTHERPANEL, NULL);
    if (bDifferenceNotFound && Opt.MessageWhenNoDiff)
    {
      const char *MsgItems[] = {
        GetMsg(MNoDiffTitle),
        GetMsg(MNoDiffBody),
        GetMsg(MOK)
      };
      Info.Message(Info.ModuleNumber,
                   0,
                   NULL,
                   MsgItems,
                   sizeof(MsgItems) / sizeof(MsgItems[0]),
                   1);
    }
  }
  // Zobrazim rozdily v editoru
  if (!bDifferenceNotFound && cResultsFileName[0]) {
		if (Opt.ShowReport) {
		  // Pokud se zobrazuje report, tak zobrazim nemodalni editor a pripadnou synchronizaci
		  // budu delat az v okamziku zavreni toho editoru. 
  		// Synchronizace se bude delat pouze u nepluginovych panelu a jen v pripade, ze se mi
	  	// podari udelat editor
		  if (Info.Editor(cResultsFileName, 
    	                GetMsg(MCmpTitle), 
                      0, 0, -1, -1, 
      	              EF_NONMODAL | EF_IMMEDIATERETURN | EF_DELETEONLYFILEONCLOSE | EF_DISABLEHISTORY,
                      0, 1) == EEC_MODIFIED) {
				EditorInfo edInfo;
		    if (!AInfo.Plugin && !PInfo.Plugin && Info.EditorControl(ECTL_GETINFO, &edInfo)) {
					// Zalozim si zaznam o tomhle editoru, abych si ho pozdeji mohl vyzvednout
					struct PendingSync * psItem;
					if (psItem = (struct PendingSync *)malloc(sizeof(*psItem))) {
						psItem->next = psEditorsQueue;
#ifndef IDENTIFY_EDITOR_BY_FILENAME
						Info.AdvControl(Info.ModuleNumber, ACTL_COMMIT, 0);
						psItem->EditorID = edInfo.EditorID;
#endif
						strcpy(psItem->LeftDir, cLeft);
						strcpy(psItem->RightDir, cRight);
						strcpy(psItem->ReportFileName, cResultsFileName);
						psEditorsQueue = psItem;
					}
				}
  	  }
		} else {
		  // Jinak muzu rovnou zobrazit dotaz na parametry synchronizace a provest to
	  	struct SyncOptions opt;
		  if (GetSyncOptions(&opt, cResultsFileName)) {
  	    Synchronize(cLeft, cRight, cResultsFileName, &opt);
    	}
		}
  }
  //    ...
  if (dwTitleSaved)
    SetConsoleTitle(cConsoleTitle);

  return INVALID_HANDLE_VALUE;
}

/****************************************************************************
 *   FAR    
 ****************************************************************************/
void WINAPI _export ExitFAR(void)
{
  if (PluginRootKey)
  {
    free(PluginRootKey);
    PluginRootKey = NULL;
  }
  struct PendingSync * q;
  while (psEditorsQueue != NULL) {
		q = psEditorsQueue->next;
		free(psEditorsQueue);
		psEditorsQueue = q;
	}
}

// Editor volam jako nemodalni, coz je sice fajn, ale potrebuju ho potom ve
// vhodnou chvili zavolat. Ta vhodna chvile je reakce na udalost EE_CLOSE
// editoru, kterym report rozdilu prohlizim.
int WINAPI _export ProcessEditorEvent(int Event, void *Param)
{
	if (Event == EE_CLOSE) {
		// Podivam se, jestli zavirany editor patri mezi ty moje
		struct PendingSync ** owner = &psEditorsQueue;
		struct PendingSync * q = psEditorsQueue;
#ifdef IDENTIFY_EDITOR_BY_FILENAME
		EditorInfo edInfo;
		if ((q != NULL) && Info.EditorControl(ECTL_GETINFO, &edInfo))	{
			while ((q != NULL) && (stricmp(edInfo.FileName, q->ReportFileName) != 0)) {
				owner = &(q->next);
				q = q->next;
			}
		} else
			q = NULL;
#else
		while ((q != NULL) && (q->EditorID != (int) Param)) {
			owner = &(q->next);
			q = q->next;
		}
#endif
		// Ano, patri (protoze jinak by to dobehlo az na konec a q by bylo NULL)
		if (q != NULL) {
			// To znamena, ze potrebuju zavolat synchronizaci
	  	struct SyncOptions opt;
		  if (GetSyncOptions(&opt, q->ReportFileName)) {
  	    Synchronize(q->LeftDir, q->RightDir, q->ReportFileName, &opt);
			}
			// A pak ten editor smazat z fronty
			*owner = q->next;
			free(q);
		}
	}
	return 0;
}


